#!/usr/bin/env python3.7

import json
import requests
from datetime import datetime, timezone
from functools import reduce
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

start = datetime.now()

##########
# Constants
##########

# labels with this color multiply by this much
color_factor = {
    'red': 2,
    'yellow': 1.5,
    'green': 0.75
}

seconds_in_day = 60 * 60 * 24
due_falloff = 10 # not a factor until this long until
due_weight = 5 # when do has this much weight added

##########
# API Communication
##########

with open(".wekan-creds.json", "r") as creds_file:
    creds = json.load(creds_file)

headers = {
  'Content-Type': 'application/x-www-form-urlencoded',
  'Accept': '*/*'
}

api_key_resp = requests.post(f"{creds['host']}/users/login", {
    'username': creds['username'],
    'password': creds['password']
}, headers = headers)

api_key = api_key_resp.json()

def wekan(url):
    return requests.get(f"{creds['host']}{url}", headers = {
        'Accept': 'application/json',
        'Authorization': f"Bearer {api_key['token']}"
    }).json()

##########
# Fetch and collect
##########

user = wekan('/api/user')

cards = []
labels = {}
boards = {}
lists = {}

for board_id in user['profile']['starredBoards']:
    board = wekan(f"/api/boards/{board_id}")
    for label in board["labels"]:
        labels[label['_id']] = label['color']
    boards[board['_id']] = board
    board_lists = wekan(f"/api/boards/{board_id}/lists")
    for l in board_lists:
        list_id = l["_id"]
        lists[list_id] = l
        list_cards = wekan(f"/api/boards/{board_id}/lists/{list_id}/cards")
        for card in list_cards:
            cards.append(wekan(f"/api/boards/{board_id}/lists/{list_id}/cards/{card['_id']}"))

##########
# Determine priority
##########

has_due = lambda card: 'dueAt' in card and card['dueAt'] is not None

def determine_weight(card):
    weights = []
    
    if has_due(card):
        now = datetime.now(timezone.utc)
        due_datetime = datetime.fromisoformat(card['dueAt'].replace("Z", "+00:00"))
        days_until = (due_datetime - now).total_seconds() / seconds_in_day
        if days_until < 0:
            weights.append({
                'cause': 'overdue',
                'constant': due_weight
            })
        elif days_until < due_falloff:
            weights.append({
                'cause': 'due soon',
                'constant': max((1 - (days_until / due_falloff)) * due_weight, due_weight)
            })

    if 'labelIds' in card:
        for label_id in card['labelIds']:
            if label_id in labels:
                color = labels[label_id]
                if color in color_factor:
                    weights.append({
                        'cause': f"{color} label",
                        'factor': color_factor[color]
                    })
                    break # only the first color matters
    
    return weights
    
def sum_weight(weight):
    constant = sum(w['constant'] for w in weight if 'constant' in w)
    factored = reduce(lambda a, b: a * b, (w['factor'] for w in weight if 'factor' in w), constant)
    
    return factored

def describe_weight(weight):
    constants = [w for w in weight if 'constant' in w]
    constants.sort(key=lambda w: w['constant'])
    constants = [f"{w['cause']}({w['constant']:.2f})" for w in constants]
    factors = [w for w in weight if 'factor' in w]
    factors.sort(key=lambda w: w['factor'])
    factors = [f"{w['cause']}({w['factor']}x)" for w in factors]

    return f"{' + '.join(constants)} {' '.join(factors)}"

priority = sorted(cards, key=lambda c: sum_weight(determine_weight(c)), reverse=True)

##########
# Send email digest
##########

with open(".smtp-creds.json", "r") as creds_file:
    smtp_creds = json.load(creds_file)

formatted_date = datetime.now().strftime('%A %Y-%m-%d')

sender = smtp_creds['sender']
receiver = smtp_creds['receiver']

msg = MIMEMultipart()
msg['Subject'] = f"Daily Task Digest for {formatted_date}"
msg['From'] = f"Task Digest <{sender}>"
msg['To'] = f"Vincent Khougaz <{receiver}>"

def format_card(card):
    weight = determine_weight(card)
    return f"""\
        <li>
            <a
                href="{creds['host']}/b/{card['boardId']}/{boards[card['boardId']]['slug']}/{card['_id']}">
                {card['title']} ({sum_weight(weight):.2f})
            </a>
            <br>
            <pre>{describe_weight(weight)}</pre>
        </li>
"""

html = f"""\
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "https://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="https://www.w3.org/1999/xhtml">
<head>
<title>Daily Task Digest for {formatted_date}</title>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
<meta http-equiv="X-UA-Compatible" content="IE=edge" />
<meta name="viewport" content="width=device-width, initial-scale=1.0 " />
</head>
<body>
Prioritized tasks for today:
<ul>
{"".join(format_card(c) for c in priority)}
</ul>
</body>
</html>
"""

msg.attach(MIMEText(html, 'html'))

smtp = smtplib.SMTP_SSL(smtp_creds['host'], smtp_creds['port'])
smtp.login(smtp_creds['username'], smtp_creds['password'])
smtp.sendmail(sender, [receiver], msg.as_string())

finish = datetime.now()

print(f"Digest sent for {formatted_date} in {(finish - start).total_seconds()}s")
